@@ -28,6 +28,7 @@ import (
 
 	"github.com/openshift/cluster-version-operator/lib"
 	"github.com/openshift/cluster-version-operator/pkg/payload"
+	"github.com/openshift/cluster-version-operator/pkg/payload/precondition"
 )
 
 func setupCVOTest(payloadDir string) (*Operator, map[string]runtime.Object, *fake.Clientset, *dynamicfake.FakeDynamicClient, func()) {
@@ -89,6 +90,7 @@ func setupCVOTest(payloadDir string) (*Operator, map[string]runtime.Object, *fak
 		wait.Backoff{
 			Steps: 1,
 		},
+		"",
 	)
 	o.configSync = worker
 
@@ -620,6 +622,267 @@ func TestCVO_StartupAndSyncUnverifiedPayload(t *testing.T) {
 	expectGet(t, actions[0], "clusterversions", "", "version")
 }
 
+func TestCVO_StartupAndSyncPreconditionFailing(t *testing.T) {
+	o, cvs, client, _, shutdownFn := setupCVOTest("testdata/payloadtest")
+
+	ctx, cancel := context.WithCancel(context.Background())
+	defer cancel()
+
+	defer shutdownFn()
+
+	worker := o.configSync.(*SyncWorker)
+	// Need the precondition check to fail permanently, so setting failure until 100 attempt to simulate that.
+	worker.preconditions = []precondition.Precondition{&testPrecondition{SuccessAfter: 100}}
+	worker.retriever.(*fakeDirectoryRetriever).Info = PayloadInfo{
+		Directory: "testdata/payloadtest",
+		Local:     true,
+	}
+	go worker.Start(ctx, 1)
+
+	// Step 1: Verify the CVO creates the initial Cluster Version object
+	//
+	client.ClearActions()
+	err := o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions := client.Actions()
+	if len(actions) != 3 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	// read from lister
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	// read before create
+	expectGet(t, actions[1], "clusterversions", "", "version")
+	// create initial version
+	actual := cvs["version"].(*configv1.ClusterVersion)
+	expectCreate(t, actions[2], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name: "version",
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID: actual.Spec.ClusterID,
+			Channel:   "fast",
+		},
+	})
+	verifyAllStatus(t, worker.StatusCh())
+
+	// Step 2: Ensure the CVO reports a status error if it has nothing to sync
+	//
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	actual = cvs["version"].(*configv1.ClusterVersion)
+	expectUpdateStatus(t, actions[1], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:       "version",
+			Generation: 1,
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID: actual.Spec.ClusterID,
+			Channel:   "fast",
+		},
+		Status: configv1.ClusterVersionStatus{
+			History: []configv1.UpdateHistory{
+				// empty because the operator release image is not set, so we have no input
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionFalse},
+				// report back to the user that we don't have enough info to proceed
+				{Type: ClusterStatusFailing, Status: configv1.ConditionTrue, Message: "No configured operator version, unable to update cluster"},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionTrue, Message: "Unable to apply <unknown>: an error occurred"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	})
+	verifyAllStatus(t, worker.StatusCh())
+
+	// Step 3: Given an operator image, begin synchronizing
+	//
+	o.releaseImage = "image/image:1"
+	o.releaseVersion = "4.0.1"
+	desired := configv1.Update{Version: "4.0.1", Image: "image/image:1"}
+	//
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	actual = cvs["version"].(*configv1.ClusterVersion)
+	expectUpdateStatus(t, actions[1], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:       "version",
+			Generation: 1,
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID: actual.Spec.ClusterID,
+			Channel:   "fast",
+		},
+		Status: configv1.ClusterVersionStatus{
+			Desired:            desired,
+			ObservedGeneration: 1,
+			History: []configv1.UpdateHistory{
+				{State: configv1.PartialUpdate, Image: "image/image:1", Version: "4.0.1", StartedTime: defaultStartedTime},
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionFalse},
+				// cleared failing status and set progressing
+				{Type: ClusterStatusFailing, Status: configv1.ConditionFalse},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionTrue, Message: "Working towards 4.0.1"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	})
+	verifyAllStatus(t, worker.StatusCh(),
+		SyncWorkerStatus{
+			Step:    "RetrievePayload",
+			Initial: true,
+			// the desired version is briefly incorrect (user provided) until we retrieve the image
+			Actual:     configv1.Update{Version: "4.0.1", Image: "image/image:1"},
+			Generation: 1,
+		},
+		SyncWorkerStatus{
+			Step:         "ApplyResources",
+			Initial:      true,
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			LastProgress: time.Unix(1, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Fraction:     float32(1) / 3,
+			Step:         "ApplyResources",
+			Initial:      true,
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			LastProgress: time.Unix(2, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Fraction:     float32(2) / 3,
+			Initial:      true,
+			Step:         "ApplyResources",
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			LastProgress: time.Unix(3, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Reconciling:  true,
+			Completed:    1,
+			Fraction:     1,
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			LastProgress: time.Unix(4, 0),
+			Generation:   1,
+		},
+	)
+
+	// Step 4: Now that sync is complete, verify status is updated to represent image contents
+	//
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	// update the status to indicate we are synced, available, and report versions
+	actual = cvs["version"].(*configv1.ClusterVersion)
+	expectUpdateStatus(t, actions[1], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:       "version",
+			Generation: 1,
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID: actual.Spec.ClusterID,
+			Channel:   "fast",
+		},
+		Status: configv1.ClusterVersionStatus{
+			ObservedGeneration: 1,
+			// Prefers the image version over the operator's version (although in general they will remain in sync)
+			Desired:     configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			VersionHash: "6GC9TkkG9PA=",
+			History: []configv1.UpdateHistory{
+				// Because image and operator had mismatched versions, we get two entries (which shouldn't happen unless there is a bug in the CVO)
+				{State: configv1.CompletedUpdate, Image: "image/image:1", Version: "1.0.0-abc", StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+				{State: configv1.PartialUpdate, Image: "image/image:1", Version: "4.0.1", StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionTrue, Message: "Done applying 1.0.0-abc"},
+				{Type: ClusterStatusFailing, Status: configv1.ConditionFalse},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionFalse, Message: "Cluster version is 1.0.0-abc"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	})
+
+	// Step 5: Wait for the SyncWorker to trigger a reconcile (500ms after the first)
+	//
+	verifyAllStatus(t, worker.StatusCh(),
+		SyncWorkerStatus{
+			Reconciling: true,
+			Step:        "ApplyResources",
+			VersionHash: "6GC9TkkG9PA=",
+			Actual:      configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			Generation:  1,
+		},
+		SyncWorkerStatus{
+			Reconciling: true,
+			Fraction:    float32(1) / 3,
+			Step:        "ApplyResources",
+			VersionHash: "6GC9TkkG9PA=",
+			Actual:      configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			Generation:  1,
+		},
+		SyncWorkerStatus{
+			Reconciling: true,
+			Fraction:    float32(2) / 3,
+			Step:        "ApplyResources",
+			VersionHash: "6GC9TkkG9PA=",
+			Actual:      configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			Generation:  1,
+		},
+		SyncWorkerStatus{
+			Reconciling:  true,
+			Completed:    2,
+			Fraction:     1,
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.0-abc", Image: "image/image:1"},
+			LastProgress: time.Unix(1, 0),
+			Generation:   1,
+		},
+	)
+
+	// Step 6: After a reconciliation, there should be no status change because the state is the same
+	//
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 1 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+}
+
 func TestCVO_UpgradeUnverifiedPayload(t *testing.T) {
 	o, cvs, client, _, shutdownFn := setupCVOTest("testdata/payloadtest-2")
 
@@ -1107,6 +1370,230 @@ func TestCVO_UpgradeUnverifiedPayloadRetriveOnce(t *testing.T) {
 	)
 }
 
+func TestCVO_UpgradePreconditionFailing(t *testing.T) {
+	o, cvs, client, _, shutdownFn := setupCVOTest("testdata/payloadtest-2")
+
+	// Setup: a successful sync from a previous run, and the operator at the same image as before
+	//
+	o.releaseImage = "image/image:0"
+	o.releaseVersion = "1.0.0-abc"
+	desired := configv1.Update{Version: "1.0.1-abc", Image: "image/image:1"}
+	uid, _ := uuid.NewRandom()
+	clusterUID := configv1.ClusterID(uid.String())
+	cvs["version"] = &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:            "version",
+			ResourceVersion: "1",
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID:     clusterUID,
+			Channel:       "fast",
+			DesiredUpdate: &desired,
+		},
+		Status: configv1.ClusterVersionStatus{
+			// Prefers the image version over the operator's version (although in general they will remain in sync)
+			Desired:     desired,
+			VersionHash: "6GC9TkkG9PA=",
+			History: []configv1.UpdateHistory{
+				{State: configv1.CompletedUpdate, Image: "image/image:0", Version: "1.0.0-abc", Verified: true, StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionTrue, Message: "Done applying 1.0.0-abc"},
+				{Type: ClusterStatusFailing, Status: configv1.ConditionFalse},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionFalse, Message: "Cluster version is 1.0.0-abc"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	}
+
+	ctx, cancel := context.WithCancel(context.Background())
+	defer cancel()
+
+	defer shutdownFn()
+
+	worker := o.configSync.(*SyncWorker)
+	worker.preconditions = []precondition.Precondition{&testPrecondition{SuccessAfter: 3}}
+
+	go worker.Start(ctx, 1)
+
+	// Step 1: The operator should report that it is blocked on precondition checks failing
+	//
+	client.ClearActions()
+	err := o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions := client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+
+	verifyAllStatus(t, worker.StatusCh(),
+		SyncWorkerStatus{
+			Step:   "RetrievePayload",
+			Actual: configv1.Update{Version: "1.0.1-abc", Image: "image/image:1"},
+		},
+		SyncWorkerStatus{
+			Step:   "PreconditionChecks",
+			Actual: configv1.Update{Version: "1.0.1-abc", Image: "image/image:1"},
+		},
+		SyncWorkerStatus{
+			Step:    "PreconditionChecks",
+			Failure: &payload.UpdateError{Reason: "UpgradePreconditionCheckFailed", Message: "Precondition \"TestPrecondition SuccessAfter: 3\" failed because of \"CheckFailure\": failing, attempt: 1 will succeed after 3 attempt", Name: "PreconditionCheck"},
+			Actual:  configv1.Update{Version: "1.0.1-abc", Image: "image/image:1"},
+		},
+	)
+
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	actual := cvs["version"].(*configv1.ClusterVersion)
+	expectUpdateStatus(t, actions[1], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:            "version",
+			ResourceVersion: "1",
+			Generation:      1,
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID:     clusterUID,
+			Channel:       "fast",
+			DesiredUpdate: &desired,
+		},
+		Status: configv1.ClusterVersionStatus{
+			// Prefers the image version over the operator's version (although in general they will remain in sync)
+			Desired:     desired,
+			VersionHash: "6GC9TkkG9PA=",
+			History: []configv1.UpdateHistory{
+				{State: configv1.PartialUpdate, Image: "image/image:1", Version: "1.0.1-abc", StartedTime: defaultStartedTime},
+				{State: configv1.CompletedUpdate, Image: "image/image:0", Version: "1.0.0-abc", Verified: true, StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionTrue, Message: "Done applying 1.0.0-abc"},
+				// cleared failing status and set progressing
+				{Type: ClusterStatusFailing, Status: configv1.ConditionTrue, Reason: "UpgradePreconditionCheckFailed", Message: "Precondition \"TestPrecondition SuccessAfter: 3\" failed because of \"CheckFailure\": failing, attempt: 1 will succeed after 3 attempt"},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionTrue, Reason: "UpgradePreconditionCheckFailed", Message: "Unable to apply 1.0.1-abc: it may not be safe to apply this update"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	})
+
+	// Step 2: Set allowUnverifiedImages to true, trigger a sync and the operator should apply the payload
+	//
+	// set an updtae
+	copied := desired
+	copied.Force = true
+	actual.Spec.DesiredUpdate = &copied
+	//
+	// ensure the sync worker tells the sync loop about it
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// wait until we see the new payload show up
+	count := 0
+	for {
+		var status SyncWorkerStatus
+		select {
+		case status = <-worker.StatusCh():
+		case <-time.After(3 * time.Second):
+			t.Fatalf("never saw expected sync event")
+		}
+		if status.Step == "RetrievePayload" && reflect.DeepEqual(configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true}, status.Actual) {
+			break
+		}
+		t.Logf("Unexpected status waiting to see first retrieve: %#v", status)
+		count++
+		if count > 8 {
+			t.Fatalf("saw too many sync events of the wrong form")
+		}
+	}
+	verifyAllStatus(t, worker.StatusCh(),
+		SyncWorkerStatus{
+			Step:         "PreconditionChecks",
+			Actual:       configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			LastProgress: time.Unix(1, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Step:         "ApplyResources",
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			LastProgress: time.Unix(2, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Fraction:     float32(1) / 3,
+			Step:         "ApplyResources",
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			LastProgress: time.Unix(3, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Fraction:     float32(2) / 3,
+			Step:         "ApplyResources",
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			LastProgress: time.Unix(4, 0),
+			Generation:   1,
+		},
+		SyncWorkerStatus{
+			Reconciling:  true,
+			Completed:    1,
+			Fraction:     1,
+			VersionHash:  "6GC9TkkG9PA=",
+			Actual:       configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			LastProgress: time.Unix(5, 0),
+			Generation:   1,
+		},
+	)
+	client.ClearActions()
+	err = o.sync(o.queueKey())
+	if err != nil {
+		t.Fatal(err)
+	}
+	actions = client.Actions()
+	if len(actions) != 2 {
+		t.Fatalf("%s", spew.Sdump(actions))
+	}
+	expectGet(t, actions[0], "clusterversions", "", "version")
+	expectUpdateStatus(t, actions[1], "clusterversions", "", &configv1.ClusterVersion{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:            "version",
+			ResourceVersion: "1",
+			Generation:      1,
+		},
+		Spec: configv1.ClusterVersionSpec{
+			ClusterID:     actual.Spec.ClusterID,
+			Channel:       "fast",
+			DesiredUpdate: &copied,
+		},
+		Status: configv1.ClusterVersionStatus{
+			ObservedGeneration: 1,
+			Desired:            configv1.Update{Version: "1.0.1-abc", Image: "image/image:1", Force: true},
+			VersionHash:        "6GC9TkkG9PA=",
+			History: []configv1.UpdateHistory{
+				{State: configv1.CompletedUpdate, Image: "image/image:1", Version: "1.0.1-abc", StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+				{State: configv1.CompletedUpdate, Image: "image/image:0", Version: "1.0.0-abc", Verified: true, StartedTime: defaultStartedTime, CompletionTime: &defaultCompletionTime},
+			},
+			Conditions: []configv1.ClusterOperatorStatusCondition{
+				{Type: configv1.OperatorAvailable, Status: configv1.ConditionTrue, Message: "Done applying 1.0.1-abc"},
+				{Type: ClusterStatusFailing, Status: configv1.ConditionFalse},
+				{Type: configv1.OperatorProgressing, Status: configv1.ConditionFalse, Message: "Cluster version is 1.0.1-abc"},
+				{Type: configv1.RetrievedUpdates, Status: configv1.ConditionFalse},
+			},
+		},
+	})
+}
+
 func TestCVO_UpgradeVerifiedPayload(t *testing.T) {
 	o, cvs, client, _, shutdownFn := setupCVOTest("testdata/payloadtest-2")
 
